import argparse
from typing import Generator, Tuple

from sentence_transformers import SentenceTransformer,  SentencesDataset, losses, models
from sentence_transformers.readers import InputExample
from torch.utils.data import DataLoader
from sentence_transformers.evaluation import BinaryClassificationEvaluator, TripletEvaluator
import logging
from random import choice, shuffle
import random

from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel, ElectionType, ElectionLocationType
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import \
    StandardProtocolSchemaRegistryFactory
import pandas as pd

BATCH_SIZE=128
EPOCHS=3
NEGATIVE_RANDOM_SAMPLES = 5
# One schema is for train, another is for validation
# TODO: add cross-validation
TRAIN_SCHEMA = (ElectionLevel.MUNICIPAL, ElectionType.REPRESENTATIVE, ElectionLocationType.CITY_RURAL)
VAL_SCHEMA = (ElectionLevel.REGIONAL, ElectionType.PERSONAL, None)

random.seed(17)

def get_anchor_name(anchor) -> str:
    return anchor["doc"].split(":")[1].strip()


def get_anchor_and_aliases(anchor: dict) -> [str]:
    return [get_anchor_name(anchor)] + anchor["aliases"]

def generate_train_data(anchors) -> Generator[Tuple[str, str, str], None, None]:
    """
    Generates train data as 3-string tuples: (anchor,positive,negative), where anchor - standardized field name, positive - its alias, negative - randomly sample of other standardized name or alias
    :param anchors:
    :return: generator of tuples (anchor,positive,negative)
    """
    for anchor in anchors:
        for alias in anchor["aliases"]:
            for _ in range(NEGATIVE_RANDOM_SAMPLES):
                random_other_anchor = choice([t for t in anchors if t["doc"] != anchor["doc"]])
                random_negative = choice(get_anchor_and_aliases(random_other_anchor))
                yield (get_anchor_name(anchor), alias, random_negative)

def run():
    parser = argparse.ArgumentParser()

    parser.add_argument('--output-model-dir',
                        dest='output_model_dir',
                        required=True,
                        type=str,
                        help='Path to save the trained model checkpoints')

    parser.add_argument('--base-model',
                        dest='base_model',
                        required=False,
                        type=str,
                        default="DeepPavlov/rubert-base-cased-sentence",
                        help='Base pre-trained model (by default, sentence encoder based on DeepPavlov RuBERT), which will be fine-tuned with triplet loss. See https://www.sbert.net/docs/pretrained_models.html')

    parser.add_argument('--triple-margin',
                        dest='triple_margin',
                        required=False,
                        default=1,
                        type=float,
                        help='Margin in triplet loss')

    args = parser.parse_args()

    schema_registry = StandardProtocolSchemaRegistryFactory.get_schema_registry()
    train_schema = schema_registry.search_schema(*TRAIN_SCHEMA)
    train_anchors = [f for f in train_schema["fields"] if 'doc' in f and f['doc'].startswith("Строка")]
    # Train data
    train_data = list(generate_train_data(train_anchors))
    pd.DataFrame(train_data).to_csv("/tmp/train.csv")
    shuffle(train_data)
    train_examples = [InputExample(texts=list(t)) for t in train_data]
    print(f"Training examples: {len(train_examples)}")
    # Model
    base_model = models.Transformer(args.base_model)
    pooling_model = models.Pooling(base_model.get_word_embedding_dimension())
    model = SentenceTransformer(modules=[base_model, pooling_model])
    
        
    train_dataset = SentencesDataset(train_examples, model)
    train_dataloader = DataLoader(train_dataset, shuffle=True, batch_size=BATCH_SIZE)
    train_loss = losses.TripletLoss(model=model, triplet_margin=args.triple_margin)
    # Validation data:
    val_schema = schema_registry.search_schema(*VAL_SCHEMA)
    val_anchors = [f for f in val_schema["fields"] if 'doc' in f and f['doc'].startswith("Строка")]
    val_data = list(generate_train_data(val_anchors))
    
    positives = [(anchor, positive, 1) for anchor, positive, _ in val_data]
    negatives = [(anchor, negative, 0) for anchor, _, negative in val_data]
    evaluation_set = positives + negatives
    print(print(f"Validation examples: {len(evaluation_set)}"))
    evaluator = BinaryClassificationEvaluator(sentences1=[s1 for s1, _,_ in evaluation_set],
                                              sentences2=[s2 for _, s2,_ in evaluation_set],
                                              labels=[l for _, _, l in evaluation_set],
                                              name="evaluation",
                                              show_progress_bar=True)

    model.fit(train_objectives=[(train_dataloader, train_loss)], epochs=EPOCHS, warmup_steps=0, evaluator=evaluator, evaluation_steps=1, show_progress_bar=True, output_path=args.output_model_dir, checkpoint_save_steps=1)


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)
    run()


from typing import Dict

from org.codeforrussia.selector.config.global_config import GlobalConfig
from org.codeforrussia.selector.standardizer.common import Standardizer
from org.codeforrussia.selector.standardizer.custom.shpilkin_format import ShpilkinDumpStandardizer
from org.codeforrussia.selector.standardizer.recognizers.protocol_field_recognizer_registry_factory import \
    ProtocolFieldRecognizerRegistryFactory
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import \
    StandardProtocolSchemaRegistryFactory

from enum import Enum

class SupportedInputFormat(Enum):
    SHPILKIN_DUMP  = "SHPILKIN"
    # TODO: add other custom formats of scraped election data

    def __str__(self):
        return self.value


class StandardizerRegistryFactory(object):

    def get_standardizer_registry(
            schema_registry_factory: StandardProtocolSchemaRegistryFactory,
            protocol_recognizer_registry_factory: ProtocolFieldRecognizerRegistryFactory,
            global_config: GlobalConfig,
    ):

        class StandardizerRegistry:
            """
            Registry of supported input data formats and their respective standardizers.
            """
            def __init__(self,
                         schema_registry_factory: StandardProtocolSchemaRegistryFactory,
                         protocol_recognizer_registry_factory: ProtocolFieldRecognizerRegistryFactory,
                         global_config: GlobalConfig,
                         ):
                self.standardizers: Dict[SupportedInputFormat, Standardizer] = {
                        SupportedInputFormat.SHPILKIN_DUMP : ShpilkinDumpStandardizer(
                            schema_registry_factory=schema_registry_factory,
                            protocol_recognizer_registry_factory=protocol_recognizer_registry_factory,
                            global_config=global_config,
                            )
                }

            def get_standardizer(self, data_format: SupportedInputFormat) -> Standardizer:
                try:
                    return self.standardizers[data_format]
                except KeyError:
                    raise NotImplementedError(f"This input data format is not supported yet: {data_format}")

        return StandardizerRegistry(schema_registry_factory,
                                    protocol_recognizer_registry_factory,
                                    global_config=GlobalConfig( # Copy global config to trigger post_init and setting environment variable
                                        gcs_bucket=global_config.gcs_bucket,
                                        ml_models_gcs_prefix=global_config.ml_models_gcs_prefix,
                                        gcs_credentials=global_config.gcs_credentials))

    get_standardizer_registry = staticmethod(get_standardizer_registry)


from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Dict, Union
from sentence_transformers import SentenceTransformer, util
import numpy as np
import logging

from torch import Tensor

from org.codeforrussia.selector.config.global_config import GlobalConfig
from org.codeforrussia.selector.utils.gcs import GCS
from org.codeforrussia.selector.utils.file import unzip

@dataclass
class ProtocolRow:
    """Represents a single row in a protocol"""
    # line number, usually integer between 1 and 12, but there can be alphanumeric value, such as "2b"
    line_number: str
    # Protocol data numbers or candidate names
    line_name: str
    # Protocol vote counts
    line_value: int
    

class ProtocolFieldRecognizer(ABC):
    PROTOCOL_FIELD_PATTERN = "Строка"

    def get_protocol_fields(self, schema) -> [str]:
        """
        Gets protocol fields for a given schema
        :param schema:
        :return:
        """
        return [f['name'] for f in schema['fields'] if 'doc' in f and f['doc'].startswith(self.PROTOCOL_FIELD_PATTERN)]

    @abstractmethod
    def recognize(self, protocol_data: List[ProtocolRow], schema: Dict) -> Dict:
        """
        Recognizes standard protocol fields in protocol rows w.r.t. standard schema
        :param protocol_data: protocol rows
        :param schema: recognized standard schema
        :return:
        """
        pass

class LineNumberBasedProtocolRecognizer(ProtocolFieldRecognizer):
    """
    Simplified recognizer that relies on line numbers. For example, federal-level election protocols are well-defined by the federal laws, no smart recognition is needed
    """
    def recognize(self, protocol_data: List[ProtocolRow], schema: Dict) -> Dict:
        standard_protocol_fields = self.get_protocol_fields(schema)
        standardized_protocol_data = {}
        for row in protocol_data:
            if row.line_name and row.line_value:
                line_number = int(row.line_number) # line number is necessarily integer
                if line_number > 0 and line_number < len(standard_protocol_fields) + 1:
                    standardized_protocol_data[standard_protocol_fields[line_number - 1]] = row.line_value
        return standardized_protocol_data

class SimilarityBasedProtocolRecognizer(ProtocolFieldRecognizer):
    """
    Recognizes protocol fields based on name similarity w.r.t. standard schema fields. There are regional electoral laws, defining the standard of protocol fields. To avoid manual work of collecting all regional  laws, we introduce a universal protocol field schema for all regions and take advantage of ML-based NLP model for recognition.

    The core ML model is loaded lazily.
    """
    def __init__(self, global_config: GlobalConfig, model_name: str):
        self.MODEL_THRESHOLD = 0.83
        self._global_config = global_config
        self._model_name = model_name

    @property
    @lru_cache()
    def model(self):
        model_file_name = f"{self._model_name}.zip"
        with TemporaryDirectory() as output_local_dir:
            output_local_filename = str(Path(output_local_dir) / model_file_name)
            logging.debug(f"Loading model='{model_file_name}' from GCS...")
            GCS().download_file(
                bucket_name=self._global_config.gcs_bucket,
                filename=str(Path(self._global_config.ml_models_gcs_prefix) / model_file_name),
                output_local_filename=output_local_filename,
            )
            logging.debug("Model downloaded")
            logging.debug("Model unzipping...")
            unzip(output_local_filename, output_local_dir)
            return SentenceTransformer(str(Path(output_local_dir) / self._model_name))

    def listToTuple(function):
        def wrapper(*args):
            args = [tuple(x) if type(x) == list else x for x in args]
            result = function(*args)
            result = tuple(result) if type(result) == list else result
            return result
        return wrapper

    @listToTuple
    @lru_cache(maxsize=200)
    def encode_sentences(self, sentences: List[str]) -> Union[List[Tensor], np.ndarray, Tensor]:
        return self.model.encode(sentences, convert_to_tensor=True)
        
    def recognize(self, protocol_data: List[ProtocolRow], schema: Dict) -> Dict:
        standard_protocol_fields = [f for f in schema['fields'] if 'doc' in f and f['doc'].startswith(self.PROTOCOL_FIELD_PATTERN)]
        standard_protocol_aliases = {a: f for f in standard_protocol_fields for a in f['aliases']}
        for f in standard_protocol_fields:
            standard_protocol_aliases[f['doc'].split(':')[1].strip()] = f

        standardized_field_names = [*standard_protocol_aliases]

        standardized_field_embeddings = self.encode_sentences(standardized_field_names)

        protocol_fields = [f.line_name for f in protocol_data]
        protocol_field_embeddings = self.encode_sentences(protocol_fields)
        cosine_scores = util.pytorch_cos_sim(protocol_field_embeddings, standardized_field_embeddings)

        standardized_protocol_data = {}
        max_score_indices = set()
        for i in range(len(protocol_fields)):
            max_score_index = np.argmax(cosine_scores[i]).item()

            max_similarity_score = cosine_scores[i][max_score_index]
            if max_similarity_score > self.MODEL_THRESHOLD:
                standard_field_id = standard_protocol_aliases[standardized_field_names[max_score_index]]['name']
                assert max_score_index not in max_score_indices, f"Model gives an invalid repetitive prediction: '{standard_field_id}' duplicate on {protocol_fields[i]} with score={max_similarity_score}"
                max_score_indices.add(max_score_index)
            
                if max_similarity_score < 1: # log the stats of non-exact matches
                    logging.debug("{}\t{} ({})\t{:.4f}".format(
                        protocol_fields[i],
                        standardized_field_names[max_score_index],
                        standard_field_id,
                        max_similarity_score))
                    
                standardized_protocol_data[standard_field_id] = protocol_data[i].line_value
            else:
                logging.debug(f"Could not recognize this protocol field: {protocol_fields[i]}. Score {max_similarity_score} below threshold = {self.MODEL_THRESHOLD}")

        return standardized_protocol_data


from typing import Tuple

from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel, ElectionLocationType, ElectionType

class ElectionAttributeRecognizer:
    def recognize(self, election_name) -> Tuple[ElectionLevel, ElectionType, ElectionLocationType]:
        """
        Recognizes election attributes from given name
        :param election_name: election name from the the protocol
        :return: tuple(election level, election type, election location type)
        """
        # TODO: add more accurate recognition; support all combinations
        lowercased_election_name = election_name.lower()
        if "депутатов государственной думы" in lowercased_election_name:
            return (ElectionLevel.FEDERAL, ElectionType.REPRESENTATIVE, None)
        elif any(all(pp in lowercased_election_name for pp in p) for p in [
            ["депутат", "город", "дум"],
            ["депутат", "город", "совет"],
            ["депутат", "город", "собран"],
        ]):
            return (ElectionLevel.MUNICIPAL, ElectionType.REPRESENTATIVE, ElectionLocationType.CITY_RURAL)
        elif any(all(pp in lowercased_election_name for pp in p) for p in [
            ["губернатор"],
            ["президент", "республик"],
            ["глав", "республик"],
            ["глав", "администрац", "област"],
        ]):
            return (ElectionLevel.REGIONAL, ElectionType.PERSONAL, None)
        elif any(all(pp in lowercased_election_name for pp in p) for p in [
            ["депутат", "совет", "республик"],
            ["депутат", "собран", "област"],
            ["депутат", "думы", "област"],
            ["депутат", "собран", "автоном", "округ"],
        ]):
            return (ElectionLevel.REGIONAL, ElectionType.REPRESENTATIVE, None)
        else:
            raise NotImplementedError(f"Cannot recognize election attributes by this election name: {election_name}")


from dataclasses import dataclass


@dataclass
class GlobalConfig:
    # Google Cloud Storage (GCS) root bucket name, where all prepared artifacts are stored, e.g. "codeforrussia-selector"
    gcs_bucket: str
    # GCS dir name, where pre-trained machine learning models are stored, e.g. "ml-models"
    ml_models_gcs_prefix: str
    # Path to GCS credentials file, being read from GOOGLE_APPLICATION_CREDENTIALS environment variable
    gcs_credentials: str = None

    def __post_init__(self):
        if self.gcs_credentials is None:
            import os
            self.gcs_credentials = os.environ["GOOGLE_APPLICATION_CREDENTIALS"]
            assert self.gcs_credentials is not None, "GOOGLE_APPLICATION_CREDENTIALS environment variable must be set"


from enum import Enum

class ElectionLevel(str, Enum):
    FEDERAL = "FEDERAL"
    REGIONAL = "REGIONAL"
    MUNICIPAL = "MUNICIPAL"


class ElectionType(str, Enum):
    PERSONAL = "PERSONAL"
    REPRESENTATIVE = "REPRESENTATIVE"


class ElectionLocationType(str, Enum):
    MUNICIPAL_DISTRICT = "MUNICIPAL_DISTRICT"
    CITY_RURAL = "CITY_RURAL"


import unittest
from org.codeforrussia.selector.config.global_config import GlobalConfig
from org.codeforrussia.selector.standardizer.custom.shpilkin_format import ShpilkinDumpStandardizer
from org.codeforrussia.selector.standardizer.recognizers.protocol_field_recognizer_registry_factory import \
    ProtocolFieldRecognizerRegistryFactory
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import StandardProtocolSchemaRegistryFactory
import jsonlines
from pathlib import Path

def read_protocol_data_from_file(filename: str):
    with jsonlines.open(Path(__file__).parent / "resources" / "shpilkin" / filename) as reader:
        for protocol in reader:
            yield protocol

class TestShpilkinDumpStandardizer(unittest.TestCase):
    state_duma_data = []
    regional_head_data  = []
    municipal_deputy_city_data = []

    @classmethod
    def setUpClass(cls):
        cls.state_duma_data = read_protocol_data_from_file("state_duma_1_0_example.jsonl")
        cls.regional_head_data = list(read_protocol_data_from_file("regional_head_1_0_example.jsonl"))
        cls.municipal_deputy_city_data = read_protocol_data_from_file("municipal_deputy_city_1_0_example.jsonl")
        cls.regional_deputy_data = read_protocol_data_from_file("regional_deputy_1_0_example.jsonl")
        global_config = GlobalConfig(
            gcs_bucket="codeforrussia-selector",
            ml_models_gcs_prefix="ml-models",
        )
        cls.standardizer = ShpilkinDumpStandardizer(
            schema_registry_factory=StandardProtocolSchemaRegistryFactory,
            protocol_recognizer_registry_factory=ProtocolFieldRecognizerRegistryFactory,
            global_config=global_config,
        )

    def test_on_state_duma(self):
        actual = TestShpilkinDumpStandardizer.standardizer.convert_batch(list(TestShpilkinDumpStandardizer.state_duma_data))
        self.assertEqual(2, len(actual), "Expected length does not match")
        self.assertEqual("TERRITORY", actual[0]["sdata"]["election"]["commission_level"])
        self.assertEqual(18476, actual[0]["sdata"]["valid"])
        self.assertEqual(1183, actual[1]["sdata"]["canceled"])

    def test_on_regional_heads(self):
        row_number = 100
        actual = TestShpilkinDumpStandardizer.standardizer.convert_batch(list(TestShpilkinDumpStandardizer.regional_head_data)[:row_number])

        self.assertEqual(row_number, len(actual), "Expected length does not match")
        self.assertEqual("AREA", actual[0]["sdata"]["election"]["commission_level"])
        self.assertEqual(0, actual[0]["sdata"]["unaccounted"])
        self.assertEqual(10, actual[0]["sdata"]["issued_to_mobile"])
        self.assertEqual(79, actual[0]["sdata"]["issued_ahead"])

    def test_on_smolensk_regional_heads(self):
        actual = TestShpilkinDumpStandardizer.standardizer.convert_batch([d for d in TestShpilkinDumpStandardizer.regional_head_data if d["loc"][0] == "Выборы Губернатора Смоленской области"][:10])
        self.assertEqual(10, len(actual), "Expected length does not match")
        self.assertEqual(216, actual[0]["sdata"]["voters"], "Number of voters must be equal to expected")
        for a in actual:
            self.assertGreater(a["sdata"]["voters"], 0, "Number of voters must be greater than 0")

    def test_on_kamchatka_regional_heads(self):
        actual = TestShpilkinDumpStandardizer.standardizer.convert_batch([d for d in TestShpilkinDumpStandardizer.regional_head_data if d["loc"][0] == "Досрочные выборы Губернатора Камчатского края"][:10])
        self.assertEqual(10, len(actual), "Expected length does not match")
        print(actual[0])
        self.assertEqual(192, actual[0]["sdata"]["issued_to_commission"], "Number of voters must be equal to expected")
        for a in actual:
            self.assertGreater(a["sdata"]["issued_to_commission"], 0, "Number of voters must be greater than 0")

    def test_on_municipal_deputy_city(self):
        row_number = 100
        actual = TestShpilkinDumpStandardizer.standardizer.convert_batch(list(TestShpilkinDumpStandardizer.municipal_deputy_city_data)[:row_number])
        
        self.assertEqual(row_number, len(actual), "Expected length does not match")
        self.assertEqual("TERRITORY", actual[0]["sdata"]["election"]["commission_level"])
        self.assertEqual(0, actual[0]["sdata"]["issued_absentee_certificates_to_territorial"])
        self.assertEqual(1473, actual[0]["sdata"]["issued_to_commission"])
        self.assertEqual(1713, actual[0]["sdata"]["issued_ahead"])

    def test_on_regional_deputy(self):
        actual = TestShpilkinDumpStandardizer.standardizer.convert_batch(list(TestShpilkinDumpStandardizer.regional_deputy_data))

        self.assertEqual(20, len(actual), "Expected length does not match")
        self.assertEqual("AREA", actual[0]["sdata"]["election"]["commission_level"])
        self.assertEqual(256, actual[0]["sdata"]["issued_to_commission"])
        self.assertEqual(4, actual[0]["sdata"]["issued_to_mobile"])
        self.assertEqual(281, actual[0]["sdata"]["in_mobile"])

        self.assertEqual("AREA", actual[19]["sdata"]["election"]["commission_level"])
        self.assertEqual(565, actual[19]["sdata"]["voters"])
        self.assertEqual(7, actual[19]["sdata"]["invalid"])
        self.assertEqual(550, actual[19]["sdata"]["received_by_commission"])
